<?php
/** 
*
* @package diceMOD
* @version $Id: dice.php v0.0.0 coppro
* @copyright (c) 2006 "coppro"
* @license http://opensource.org/licenses/gpl-license.php GNU Public License 
*
*/

/**
*/
if (!defined('IN_PHPBB'))
{
	exit;
}

// Ensure that if we're being called from a function, we still can access the global user variable.
global $user;
$user->add_lang('mods/dice');

/**
* Check whether a BBCode bitfield contains a dice tag.
*/
function has_dice($bitfield)
{
	$bitfield = new bitfield($bitfield);
	return $bitfield->get(DICE_BBCODE_ID);
}

/**
* Dice roller class.
* Used after second bbcode pass to get dice rolls.
*/
class dice
{
	var $type = '';
	var $id = 0;
	var $templates = array();

	var $roll_data = array();
	var $new_data  = array();

	/**
	* Constructor.
	* Requires a type and an id.
	*/
	function dice($type = '', $id = '')
	{
		$this->type = $type;
		$this->id = $id;
		$this->prepare_template();
		$this->load_roll_data();
	}

	/**
	* Parse dice rolls in a given set of text.
	*/
	function parse(&$message)
	{
		$indices = array();
		while (preg_match('#<!--READY-->\[dice(?:=(.+))?\]#U', $message, $matches, PREG_OFFSET_CAPTURE))
		{
			$before_pos = $matches[0][1];
			$before_length = utf8_strlen($matches[0][0]);

			if (count($matches) == 2)
			{
				$title = $matches[1][0];
			}
			else
			{
				global $user;
				$title = $user->lang['ROLLS'];
			}

			$after_pos = utf8_strpos($message, '[/dice]<!--READY-->');
			if ($after_pos === false)
			{
				trigger_error('NO_DICE', E_USER_ERROR);
			}
			$rolled = false;
			
			$roll_text = '';

			preg_match_all('#(?:,|^)\s*(?:([^:]+):\s*)?(?:([0-9]+)\s*d\s*)?([0-9]+)(?:\s*%\s*([0-9]))?((?:\s*(?:[-+*/^<>]|&[lg]t;)\s*-?[0-9]+(?:\.[0-9]+)?)*)#',
					utf8_substr($message, $before_pos + $before_length, $after_pos - $before_pos - $before_length), $rolls, PREG_SET_ORDER);
			foreach ($rolls as $roll)
			{
				if (!(int) $roll[3])
				{
					continue;
				}

				$name = ($roll[1])?str_replace('\n', ' ', $roll[1]):'';

				$sides = (int) $roll[3];
				$number = ((int) $roll[2])?((int) $roll[2]):1;

				$drops = (int) $roll[4];
				if ($drops >= $number)
				{
					continue;
				}

				$mods = $roll[5];

				$rolled = true;
				$results = array();

				if (!isset($indices[$sides]))
				{
					$indices[$sides] = 0;
				}

				for ($i = 0; $i < $number; ++$i)
				{
					$value =  $this->roll($sides, $indices[$sides]);
					$results[] = $value;
					$indices[$sides]++;
				}

				$result_text = implode($results, ', ');

				$sum = array_sum($results);

				if ($drops)
				{
					global $user;
					sort($results, SORT_REGULAR);
					$dropped = array_chunk($results, $drops);
					$sum -= array_sum($dropped[0]);
					$dropping_text = implode(', ', $dropped[0]);
				}
				else
				{
					$dropping_text = '';
				}

				$mod_text = '';
				if ($mods)
				{
					preg_match_all('#((?:[-+*/^<>]|&[lg]t;))\s*(-?[0-9]+(?:\.[0-9]+)?)#', $mods, $matches, PREG_SET_ORDER);
					foreach ($matches as $match)
					{
						$value = (float) $match[2];
						if (!$value)
						{
							continue;
						}

						switch ($match[1])
						{
							case '-':
								$sum -= $value;
								break;
							case '+':
								$sum += $value;
								break;
							case '*':
								$sum *= $value;
								break;
							case '/':
								$sum /= $value;
								break;
							case '^':
								$sum = pow(abs($sum), $value);
								break;
							case '<': case '&lt;':
								if ($sum > $value)
								{
									$sum = $value;
								}
								break;
							case '>': case '&gt;':
								if ($sum < $value)
								{
									$sum = $value;
								}
								break;
							default:
								continue;
						}
						$mod_text .= ' ' . $match[1] . $match[2];
					}
				}

				$roll_text .= str_replace(array('{NAME}', '{NUMBER}', '{SIDES}', '{MODS}', '{TOTAL}', '{ROLLS}', '{DROPS}', '{DROPPED_NUMS}'),
						array($name, $number, $sides, $mod_text, $sum, $result_text, $drops, $dropping_text),
						($drops)?$this->templates['dice_drop']:$this->templates['dice_roll']);
			}

			if ($rolled)
			{
				$message = utf8_substr($message, 0, $before_pos) . str_replace('{TITLE}', $title, $this->templates['dice_begin']) . $roll_text .
							$this->templates['dice_end'] . utf8_substr($message, $after_pos + 19);
			}
			else
			{
				$message = utf8_substr($message, 0, $before_pos) . utf8_substr($message, $after_pos + 19);
			}
		}

		$this->commit_data_changes();
	}

	/**
	* Preform a roll, storing or loading from the database as applicable.
	*/
	function roll($sides, $index)
	{
		global $db;
		if (!$this->type)
		{
			return mt_rand(1, $sides);
		}

		if (array_key_exists($sides, $this->roll_data) &&
			array_key_exists($index, $this->roll_data[$sides]))
		{
			$result = $this->roll_data[$sides][$index];
		}
		else
		{
			$result = false;
		}

		if ($result === false)
		{
			$result = mt_rand(1, $sides);

			if (!isset($this->new_data[$sides]))
			{
				$this->new_data[$sides] = array();
			}

			$this->new_data[$sides][$index] = $result;
		}

		return $result;
	}

	/**
	* Prepare the template values for substitution.
	*/
	function prepare_template()
	{
		global $user, $phpbb_root_path;
		// Hardcode defaults
		$this->templates['dice_begin']	= "<div style='border-style: solid; border-width: 1px; border-color: black; margin: 3px 0px 3px 3%; padding: 3px 6px 3px 6px'> {L_ROLLS}: <div style='margin-left: 3%'>";
		$this->templates['dice_roll']	= "{NUMBER}d{SIDES}{MODS}: {TOTAL} ({ROLLS})<br />";
		$this->templates['dice_drop']   = "{NUMBER}d{SIDES}{MODS} {L_DROPPING} {DROPS}: {TOTAL} ({ROLLS} {L_DROP} {DROPPED_NUMS})<br />";
		$this->templates['dice_end']	= "</div></div>";

		$template_filename = $phpbb_root_path . 'styles/' . $user->theme['template_path'] . '/template/dice.html';

		if (file_exists($template_filename))
		{
			$tpl = file_get_contents($template_filename);
		}
		else
		{
			$tpl = false;
		}

		if ($tpl !== false)
		{
			// replace \ with \\ and then ' with \'.
			$tpl = str_replace('\\', '\\\\', $tpl);
			$tpl = str_replace("'", "\'", $tpl);

			// strip newlines and indent
			$tpl = preg_replace("/\n[\n\r\s\t]*/", '', $tpl);

			$matches = preg_match_all('#<!-- BEGIN (.*?) -->(.*?)<!-- END (?:.*?) -->#', $tpl, $match);

			for ($i = 0; $i < $matches; $i++)
			{
				$name = $match[1][$i];
				if ($name == 'dice_begin' || $name == 'dice_roll' || $name == 'dice_end' || $name == 'dice_drop')
				{
					$this->templates[$name] = $match[2][$i];
				}
			}
		}

		foreach($this->templates as $index => $text)
		{
			$this->templates[$index] = preg_replace('/{L_([A-Z_]+)}/e', "(!empty(\$user->lang['\$1'])) ? \$user->lang['\$1'] : ucwords(strtolower(str_replace('_', ' ', '\$1')))", $text);
		}
	}

	function load_roll_data()
	{
		global $db;

		$this->roll_data = array();

		if (!($this->type && $this->id))
		{
			return;
		}

		$sql_ary = array(
			'type'	=> $this->type,
			'id'	=> $this->id
		);

		$result = $db->sql_query('SELECT sides, text_index, result FROM ' . ROLLS_TABLE . ' WHERE ' . $db->sql_build_array('SELECT', $sql_ary));

		while ($row = $db->sql_fetchrow($result))
		{
			if (!isset($this->roll_data[$row['sides']]))
			{
				$this->roll_data[$row['sides']] = array();
			}
			$this->roll_data[$row['sides']][$row['text_index']] = $row['result'];
		}

		$db->sql_freeresult($result);
	}

	function commit_data_changes()
	{
		global $db;

		$sql_ary = array();

		foreach ($this->new_data as $sides => $ary)
		{
			foreach ($ary as $index => $result)
			{
				$sql_ary[] = array(
					'type'		=> $this->type,
					'id'		=> $this->id,
					'sides'		=> $sides,
					'text_index'=> $index,
					'result'	=> $result
				);
			}
		}

		if (count($sql_ary))
		{
			$db->sql_multi_insert(ROLLS_TABLE, $sql_ary);
		}

		$this->new_data = array();
	}
}